from typing import Iterator, Iterable, Self, Tuple, Hashable
from abc import ABC, abstractmethod

def CartesianProduct(E: Iterable, n: int) -> Iterator[Tuple]:
    if n == 0:
        yield ()
    else:
        yield from ((a, *e) for a in E for e in CartesianProduct(E, n-1))
        


class Relation:
    """
    Ex:
    ===
        >>> A = Relation(3)

    """

    def __init__(self, arity: int):
        self.arity = arity


class Query(ABC):
    @abstractmethod
    def free_variables(self) -> tuple[str, ...]:
        ...

    @abstractmethod
    def is_boolean(self) -> bool:
        ...


class Atom(Query):
    """
    Ex:
    ===
        >>> A = Relation(3)
        >>> query = Atom(A, "x", "y", "x")
        # A(x, y, x)
        >>> query.free_variables()
        (x, y)

    """

    def __init__(self, relation: Relation, variables: Iterable[str]):
        self.relation = relation
        self.variables = tuple(variables)

    def free_variables(self) -> Tuple[str, ...]:
        L = list()
        S = set()
        for var in self.variables:
            if var not in S:
                L.append(var)
                S.add(var)
        return tuple(L)

    def is_boolean(self) -> bool:
        return self.relation.arity == 0


class CQ(Query):
    """
    Ex:
    ===
        >>> A = Relation(3)
        >>> B = Relation(2)
        >>> query = CQ(("x", "y"), (Atome(A, "x", "x", "x"), Atom(A, "x", "y", "x"), Atom(B, "y", "y")))
        # Exists x, y, A(x, x, x) AND A(x, y, x) AND B(y, y)
        >>> query = CQ(("y", ), (Atome(A, "x", "x", "x"), Atom(A, "x", "y", "x"), Atom(B, "y", "y")))
        # Exists y, A(x, x, x) AND A(x, y, x) AND B(y, y)

    """

    def __init__(self, quantified_variables: Iterable[str], atoms: Iterable[Atom]):
        self.atoms = tuple(atoms)
        if len(self.atoms) != len(set(self.atoms)):
            raise ValueError("Quantified variable should appears only once")

        self.quantified_variables = tuple(quantified_variables)

    def is_boolean(self) -> bool:
        return bool(self.free_variables())

    def free_variables(self) -> tuple[str, ...]:
        L = []
        S = set(self.quantified_variables)
        SL = set()
        for atom in self.atoms:
            for variable in atom.variables:
                if variable not in S:
                    if variable not in SL:
                        L.append(variable)
                        SL.add(variable)
        return tuple(L)

    def is_sat(self) -> bool:
        ...

    def imply(self, other: Query):
        # TODO


class UCQ(Query):
    """
    Ex:
    ===
        >>> A = Relation(3)
        >>> B = Relation(2)
        >>> query1 = CQ(("x", "y"), (Atome(A, "x", "x", "x"), Atom(A, "x", "y", "x"), Atom(B, "y", "y")))
        # Exists x, y, A(x, x, x) AND A(x, y, x) AND B(y, y)
        >>> query2 = CQ(("y", ), (Atome(A, "x", "x", "x"), Atom(A, "x", "y", "x"), Atom(B, "y", "y")))
        # Exists y, A(x, x, x) AND A(x, y, x) AND B(y, y)
        >>> query = UCQ(query1, query2, query1)

    """

    def __init__(self, *cqs: CQ):
        self.cqs = tuple(cqs)

    def is_boolean(self) -> bool:
        # TODO
        ...

    def free_variables(self):
        # TODO
        ...


class ConcreteRelation:
    """
    Ex:
    ===
        >>> A = Relation(2)
        >>> ConctreA = ConcreteRelation(A, ((12, 32), ('a', 'b'), ((0, 1), 2)))
    """

    def __init__(self, relation: Relation, values: Iterable[tuple[Hashable, ...]]):
        values = set(values)
        values_arity = set(map(len, values))
        assert values_arity == set((relation.arity,))
        self.values = values
        self.relation = relation


class Database:
    def __init__(self, *relations: ConcreteRelation):
        self.concrete_relations = { relation_concrete.relation : relation_concrete for relation_concrete in ConcreteRelation }

    def domain(self) -> set:
        S = set()
        for relation in self.concrete_relations:
            # S.update(*relation.values)
            for val in relation.values:
                S.update(val)
        return S

    def evaluate(self, q: Query) -> Iterator[Tuple]:
        def _evaluate(q: Query, Context: dict[str, Hashable]) -> bool:
            """
            Context should map each free variable of q over a concrete value.
            """
            free_var = q.free_variables()
            if isinstance(q, Atom):
                # A(x, y, x)
                for x in free_var:
                    if x not in Context:
                        raise ValueError(f"Cannot evaluate a query without value for {x}")
                L = []
                for x in q.variables:
                    L.append(Context[x])
                t = tuple(L)
                # t is a concrete tuple
                return t in self.concrete_relations[q.relation].values
                # We build the tuple and check that it is in the relation.

            elif isinstance(q, CQ):
                # Exists x, A(x, y, x) and B(x, y)
                # y is a free variable => we should have Context[y] set to a concrete value
                # x is bounded 
                n = len(q.quantified_variables)
                possible_values = CartesianProduct(self.domain(), n) # we need to implement this

                for v in possible_values:
                    new_context = dict(Context)
                    new_context.update(zip(q.quantified_variables, v))
                    if all(_evaluate(atom, new_context) for atom in q.atoms):
                        return True

                return False
            else:
                assert isinstance(q, UCQ)
                ... 
                    
                
        # _evaluate map a query and context to boolean

        n = len(q.quantified_variables)
        possible_values = CartesianProduct(self.domain(), n) # we need to implement this
        for v in possible_values:
            context = dict(zip(q.quantified_variables, v))
            if _evaluate(q, context):
                yield v
            

    @classmethod
    def canonical_database(cls, query: Query) -> Self:
        # TODO
        ...
